using System;
using System.Diagnostics;
using System.Collections;
using System.Threading;
using Bespoke.Robotics.Common;
using Bespoke.Robotics.BrainStem.Exceptions;

namespace Bespoke.Robotics.BrainStem
{
    /// <summary>
    /// Manager for BrainStem commands.
    /// </summary>
    public sealed class CommandManager
    {
		/// <summary>
		/// Gets the initialization status of the command manager.
		/// </summary>
		/// <value>The initialization status of the command manager.</value>
		public static bool IsInitialized
		{
			get
			{
				return Instance.mIsInitialized;
			}
		}

		/// <summary>
		/// Configure a digital IO pin.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="configuration">The new configuration of the digital IO pin.</param>
		/// <exception cref="ArgumentNullException">Thrown if <paramref name="configuration"/> is null.</exception>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static void ConfigureDigitalIoPin(byte iicAddress, DigitalIoPinConfiguration configuration)
		{
			try
			{
				Verify.ParamIsNotNull("configuration", configuration);

				byte[] data = new byte[2] { (byte)configuration.DigitalIoPin, configuration.ToDataByte() };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.ConfigureDigitalIo, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.ConfigureDigitalIo, "Error configuring digital IO pin.", Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Configure the general parameters of a servo.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="configuration">The new configuration of the servo.</param>
		/// <exception cref="ArgumentNullException">Thrown if <paramref name="configuration"/> is null.</exception>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static void ConfigureServo(byte iicAddress, ServoConfiguration configuration)
		{
			try
			{
				Verify.ParamIsNotNull("configuration", configuration);

				byte[] data = new byte[2] { (byte)configuration.ServoId, configuration.ToDataByte() };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.ConfigureServo, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.ConfigureServo, "Error configuring servo.", Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Configure the starting (zero) position and resolution (range) of a servo.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="servoId">The ID of the servo to configure.</param>
		/// <param name="positionZero">Position zero for the servo (Range: 0-70).</param>
		/// <param name="positionResolution">Positioning resolution (range of travel) for the servo (Range: 1-70)</param>
		/// <exception cref="ArgumentOutOfRangeException">Thrown if <paramref name="positionZero"/> or <paramref name="positionResolution"/>
		/// are out of range.</exception>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static void ConfigureServoMotion(byte iicAddress, ServoIdType servoId, byte positionZero, byte positionResolution)
		{
			try
			{
				if (positionZero > MAX_SERVO_POSITION_ZERO)
				{
					throw new ArgumentOutOfRangeException("positionZero", positionZero, string.Format("Valid range between 0-{0}", MAX_SERVO_POSITION_ZERO));
				}

				if ((positionResolution < MIN_SERVO_POSITION_RESOLUTION) || (positionResolution > MAX_SERVO_POSITION_RESOLUTION))
				{
					throw new ArgumentOutOfRangeException("positionResolution", positionResolution, string.Format("Valid range between {0}-{1}", MIN_SERVO_POSITION_RESOLUTION, MAX_SERVO_POSITION_RESOLUTION));
				}

				byte[] data = new byte[3] { (byte)servoId, positionZero, positionResolution };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.ConfigureServoMotion, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.ConfigureServoMotion, "Error configuring servo motion.", Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

        /// <summary>
        /// Echo an array of bytes.
        /// </summary>
        /// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
        /// <param name="data">The array of raw bytes to be sent to the BrainStem.</param>
        /// <returns>A byte array containing the same bytes that were passed in.</returns>
        /// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
        public static byte[] Debug(byte iicAddress, byte[] data)
        {
			try
			{
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.Debug, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Read the reply packet.
				Packet replyPacket = Instance.GetPacketFromQueue(Packet.CommandCodeType.Debug);
				if (replyPacket == null)
				{
					// Check the error queue.
					if (Instance.mErrorQueue.Count != 0)
					{
						throw new CommandException(Packet.CommandCodeType.Debug, Instance.mErrorQueue[0]);
					}

					throw new CommandException(Packet.CommandCodeType.Debug, "Expected debug message response.");
				}

				return replyPacket.Data;
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
        }

		/// <summary>
        /// Retrieve a digital IO pin configuration.
        /// </summary>
        /// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="digitalIoPin">The digital IO pin configuration to retrieve.</param>
        /// <returns>The configuration for the specified digital IO pin.</returns>
        /// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static DigitalIoPinConfiguration GetDigitalIoPinConfiguration(byte iicAddress, DigitalIoPinType digitalIoPin)
		{
			try
			{
				byte[] data = new byte[1] { (byte)digitalIoPin };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.ConfigureDigitalIo, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Read the reply packet.
				Packet replyPacket = Instance.GetPacketFromQueue(Packet.CommandCodeType.ConfigureDigitalIo);
				if (replyPacket == null)
				{
					// Check the error queue.
					if (Instance.mErrorQueue.Count != 0)
					{
						throw new CommandException(Packet.CommandCodeType.ConfigureDigitalIo, Instance.mErrorQueue[0]);
					}

					throw new CommandException(Packet.CommandCodeType.ConfigureDigitalIo, "Expected digital IO pin configuration response.");
				}

				if ((replyPacket.Data == null) || (replyPacket.Data.Length != 2))
				{
					throw new CommandException(Packet.CommandCodeType.ConfigureDigitalIo, "Invalid digital IO pin configuration response.");
				}

				DigitalIoPinType replyPin = (DigitalIoPinType)replyPacket.Data[0];
				if (replyPin != digitalIoPin)
				{
					throw new CommandException(Packet.CommandCodeType.ConfigureDigitalIo, "Response pin did not match the requested pin.");
				}

				DigitalIoPinConfigurationFlags configurationFlags = (DigitalIoPinConfigurationFlags)replyPacket.Data[1];
				return new DigitalIoPinConfiguration(digitalIoPin, configurationFlags);
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Retrieve a servo configuration.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="servoId">The servo configuration to retrieve.</param>
		/// <returns>The configuration for the specified servo.</returns>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static ServoConfiguration GetServoConfiguration(byte iicAddress, ServoIdType servoId)
		{
			try
			{
				byte[] data = new byte[1] { (byte)servoId };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.ConfigureServo, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Read the reply packet.
				Packet replyPacket = Instance.GetPacketFromQueue(Packet.CommandCodeType.ConfigureServo);
				if (replyPacket == null)
				{
					// Check the error queue.
					if (Instance.mErrorQueue.Count != 0)
					{
						throw new CommandException(Packet.CommandCodeType.ConfigureServo, Instance.mErrorQueue[0]);
					}

					throw new CommandException(Packet.CommandCodeType.ConfigureServo, "Expected servo configuration response.");
				}

				if ((replyPacket.Data == null) || (replyPacket.Data.Length != 2))
				{
					throw new CommandException(Packet.CommandCodeType.ConfigureServo, "Invalid servo configuration response.");
				}

				ServoIdType replyServoId = (ServoIdType)replyPacket.Data[0];
				if (replyServoId != servoId)
				{
					throw new CommandException(Packet.CommandCodeType.ConfigureServo, "Response ID did not match the requested ID.");
				}

				ServoConfigurationFlags configurationFlags = (ServoConfigurationFlags)replyPacket.Data[1];
				return new ServoConfiguration(servoId, configurationFlags);
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Retrieve the starting (zero) position and resolution (range) of a servo.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="servoId">The ID of the servo to configure.</param>
		/// <param name="positionZero">Position zero for the servo (Range: 0-70).</param>
		/// <param name="positionResolution">Positioning resolution (range of travel) for the servo (Range: 1-70)</param>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static void GetServoMotion(byte iicAddress, ServoIdType servoId, out byte positionZero, out byte positionResolution)
		{
			try
			{
				byte[] data = new byte[1] { (byte)servoId };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.ConfigureServoMotion, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Read the reply packet.
				Packet replyPacket = Instance.GetPacketFromQueue(Packet.CommandCodeType.ConfigureServoMotion);
				if (replyPacket == null)
				{
					// Check the error queue.
					if (Instance.mErrorQueue.Count != 0)
					{
						throw new CommandException(Packet.CommandCodeType.ConfigureServoMotion, Instance.mErrorQueue[0]);
					}

					throw new CommandException(Packet.CommandCodeType.ConfigureServoMotion, "Expected servo motion response.");
				}

				if ((replyPacket.Data == null) || (replyPacket.Data.Length != 3))
				{
					throw new CommandException(Packet.CommandCodeType.ConfigureServo, "Invalid servo motion response.");
				}

				ServoIdType replyServoId = (ServoIdType)replyPacket.Data[0];
				if (replyServoId != servoId)
				{
					throw new CommandException(Packet.CommandCodeType.ConfigureServo, "Response ID did not match the requested ID.");
				}

				positionZero = replyPacket.Data[1];
				positionResolution = replyPacket.Data[2];
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Retrieve the absolute position of a servo.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="servoId">The ID of the servo to send the command to.</param>
		/// <returns>The absolute position of the servo (Range: 0-255).</returns>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static byte GetServoPosition(byte iicAddress, ServoIdType servoId)
		{
			try
			{
				byte[] data = new byte[1] { (byte)servoId };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.PositionServoAbsolute, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Read the reply packet.
				Packet replyPacket = Instance.GetPacketFromQueue(Packet.CommandCodeType.PositionServoAbsolute);
				if (replyPacket == null)
				{
					// Check the error queue.
					if (Instance.mErrorQueue.Count != 0)
					{
						throw new CommandException(Packet.CommandCodeType.PositionServoAbsolute, Instance.mErrorQueue[0]);
					}

					throw new CommandException(Packet.CommandCodeType.PositionServoAbsolute, "Expected servo position response.");
				}

				if ((replyPacket.Data == null) || (replyPacket.Data.Length != 2))
				{
					throw new CommandException(Packet.CommandCodeType.PositionServoAbsolute, "Invalid servo position response.");
				}

				ServoIdType replyServoId = (ServoIdType)replyPacket.Data[0];
				if (replyServoId != servoId)
				{
					throw new CommandException(Packet.CommandCodeType.PositionServoAbsolute, "Response ID did not match the requested ID.");
				}

				return replyPacket.Data[1];				
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

        /// <summary>
        /// Retrieve a system parameter.
        /// </summary>
        /// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
        /// <param name="systemParameter">The system parameter to retrieve.</param>
        /// <returns>The value of the requested system parameter.</returns>
        /// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
        public static string GetSystemParameter(byte iicAddress, SystemParameterType systemParameter)
        {
			try
			{
				// Send the command packet.
				byte[] data = new byte[1] { (byte)systemParameter };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.GetSystemParameter, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Read the reply packet.
				Packet replyPacket = Instance.GetPacketFromQueue(Packet.CommandCodeType.GetSystemParameterReply);
				if (replyPacket == null)
				{
					// Check the error queue.
					if (Instance.mErrorQueue.Count != 0)
					{
						throw new CommandException(Packet.CommandCodeType.GetSystemParameter, Instance.mErrorQueue[0]);
					}

					throw new CommandException(Packet.CommandCodeType.GetSystemParameter, "Expected system parameter response.");
				}

				// The reply packet should contain at least 3 bytes (the command code, the system parameter
				// and a 1 or 2 byte value).
				if (replyPacket.Length < 3)
				{
					throw new CommandException(Packet.CommandCodeType.GetSystemParameter, "Invalid system parameter response.");
				}

				// The reply should contain the same system parameter as was requested.
				SystemParameterType replyParameter = (SystemParameterType)replyPacket.Data[0];
				if (replyParameter != systemParameter)
				{
					throw new CommandException(Packet.CommandCodeType.GetSystemParameter, "Response parameter did not match the requested parameter.");
				}

				int dataLength = replyPacket.Data.Length - 1;
				byte[] buffer = new byte[dataLength];
				Array.Copy(replyPacket.Data, 1, buffer, 0, dataLength);
				return BitConverter.ToString(buffer);
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
        }

        /// <summary>
        /// Initialize the command manager.
        /// </summary>
        /// <param name="stream">An open stream to an Acroname BrainStem.</param>
        /// <param name="commandReplyTimeout">Time (in milliseconds) a command will with for a reply before failing.</param>
        /// <exception cref="ArgumentNullException">Thrown if <paramref name="stream"/> is null.</exception>
		/// <exception cref="CommandManagerAlreadyInitializedException">Thrown if the command manager is already initialized.</exception>
        /// <exception cref="StreamClosedException">Thrown if the stream is closed.</exception>
        public static void Init(Stream stream, int commandReplyTimeout)
        {
            Verify.ParamIsNotNull("stream", stream);

			if (Instance.mIsInitialized)
			{
				throw new CommandManagerAlreadyInitializedException();
			}

            if (stream.IsOpen == false)
            {
                throw new StreamClosedException();
            }

            Instance.mStream = stream;
            Instance.mStream.HeartbeatReceived += new HeartbeatReceivedEventHandler(mStream_HeartbeatReceived);
            Instance.mStream.PacketReceived += new PacketReceivedEventHandler(mStream_PacketReceived);
            Instance.mCommandReplyTimeout = commandReplyTimeout;
			Instance.mIsInitialized = true;
        }

		/// <summary>
		/// Read an input bit from an A2D IO pin.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="a2dIoPin">The digital IO pin to read the input from.</param>
		/// <param name="sendReplyToHost">The BrainStem will send a reply packet to the host.</param>
		/// <param name="sendReplyToReflex">The value will go directly to a reflex routine within the BrainStem.</param>
		/// <returns>The input from the specified digital IO pin.</returns>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static int ReadA2DInput(byte iicAddress, AnalogToDigitalIoPinType a2dIoPin, bool sendReplyToHost, bool sendReplyToReflex)
		{
			try
			{
				int a2dInput = 0;
				byte devCodeByte = (byte)a2dIoPin;
				if (sendReplyToHost)
				{
					devCodeByte |= (byte)DeviceValueFlags.Host;
				}

				if (sendReplyToReflex)
				{
					devCodeByte |= (byte)DeviceValueFlags.Reflex;
				}

				byte[] data = new byte[1] { (byte)devCodeByte };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.ReadA2D, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				if (sendReplyToHost)
				{
					// Read the reply packet.
					Packet replyPacket = Instance.GetPacketFromQueue(Packet.CommandCodeType.DeviceValue);
					if (replyPacket == null)
					{
						// Check the error queue.
						if (Instance.mErrorQueue.Count != 0)
						{
							throw new CommandException(Packet.CommandCodeType.ReadA2D, Instance.mErrorQueue[0]);
						}

						throw new CommandException(Packet.CommandCodeType.ReadA2D, "Expected A2D input response.");
					}

					if ((replyPacket.Data == null) || (replyPacket.Data.Length != 3))
					{
						throw new CommandException(Packet.CommandCodeType.ReadA2D, "Invalid A2D input response.");
					}

					byte replyDeviceId;
					bool replySendReplyToHost, replySendReplyToReflex;
					Instance.ParseDeviceInputCode(replyPacket.Data[0], out replyDeviceId, out replySendReplyToHost, out replySendReplyToReflex);
					AnalogToDigitalIoPinType replyA2DIoPin = (AnalogToDigitalIoPinType)replyDeviceId;
					if (replyA2DIoPin != a2dIoPin)
					{
						throw new CommandException(Packet.CommandCodeType.DigitalIo, "Response ID did not match the requested ID.");
					}

					//TODO: I don't think this is correct.
					a2dInput = replyPacket.Data[1] << 2;
					a2dInput += replyPacket.Data[2];
					if (a2dInput > MAX_A2D_VALUE)
					{
						a2dInput = MAX_A2D_VALUE;
					}
				}

				return a2dInput;
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Read an input bit from a digital IO pin.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="digitalIoPin">The digital IO pin to read the input from.</param>
		/// <param name="sendReplyToHost">The BrainStem will send a reply packet to the host.</param>
		/// <param name="sendReplyToReflex">The value will go directly to a reflex routine within the BrainStem.</param>
		/// <returns>The input from the specified digital IO pin.</returns>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static bool ReadDigitalInput(byte iicAddress, DigitalIoPinType digitalIoPin, bool sendReplyToHost, bool sendReplyToReflex)
		{
			try
			{
				bool digitalInput = false;
				byte devCodeByte = (byte)digitalIoPin;
				if (sendReplyToHost)
				{
					devCodeByte |= (byte)DeviceValueFlags.Host;
				}

				if (sendReplyToReflex)
				{
					devCodeByte |= (byte)DeviceValueFlags.Reflex;
				}

				byte[] data = new byte[1] { (byte)devCodeByte };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.DigitalIo, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				if (sendReplyToHost)
				{
					// Read the reply packet.
					Packet replyPacket = Instance.GetPacketFromQueue(Packet.CommandCodeType.DeviceValue);
					if (replyPacket == null)
					{
						// Check the error queue.
						if (Instance.mErrorQueue.Count != 0)
						{
							throw new CommandException(Packet.CommandCodeType.DigitalIo, Instance.mErrorQueue[0]);
						}

						throw new CommandException(Packet.CommandCodeType.DigitalIo, "Expected digital input response.");
					}

					if ((replyPacket.Data == null) || (replyPacket.Data.Length != 2))
					{
						throw new CommandException(Packet.CommandCodeType.DigitalIo, "Invalid digital input response.");
					}

					byte replyDeviceId;
					bool replySendReplyToHost, replySendReplyToReflex;
					Instance.ParseDeviceInputCode(replyPacket.Data[0], out replyDeviceId, out replySendReplyToHost, out replySendReplyToReflex);
					DigitalIoPinType replyDigitalIoPin = (DigitalIoPinType)replyDeviceId;
					if (replyDigitalIoPin != digitalIoPin)
					{
						throw new CommandException(Packet.CommandCodeType.DigitalIo, "Response ID did not match the requested ID.");
					}

					digitalInput = Convert.ToBoolean(replyPacket.Data[1]);
				}

				return digitalInput;
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Reset a digital IO pin.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="digitalIoPin">The digital IO pin to reset.</param>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static void ResetDigitalIoPin(byte iicAddress, DigitalIoPinType digitalIoPin)
		{
			try
			{
				byte[] data = new byte[1] { (byte)digitalIoPin };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.ResetDigitalIo, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.ResetDigitalIo, Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

        /// <summary>
        /// Save the servo parameters to the BrainStem's EEPROM.
        /// </summary>
        /// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
        /// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
        public static void SaveServoParameters(byte iicAddress)
        {
			try
			{
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.SaveServoParameters, Packet.DirectionType.HostToStem);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.SaveServoParameters, Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
        }

        /// <summary>
        /// Save the system parameters to the BrainStem's EEPROM.
        /// </summary>
        /// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
        /// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
        public static void SaveSystemParameters(byte iicAddress)
        {
			try
			{
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.SaveSystemParameters, Packet.DirectionType.HostToStem);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.SaveSystemParameters, Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
        }

		/// <summary>
		/// Sets the absolute position of a servo.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="servoId">The ID of the servo to send the command to.</param>
		/// <param name="absolutePosition">The absolute position of the servo (Range: 0-255).</param>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static void SetServoPositionAbsolute(byte iicAddress, ServoIdType servoId, byte absolutePosition)
		{
			try
			{
				byte[] data = new byte[2] { (byte)servoId, absolutePosition };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.PositionServoAbsolute, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.PositionServoAbsolute, Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Sets the relative position of a servo by adding or subtracting an offset to the servo's
		/// current position.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="servoId">The ID of the servo to send the command to.</param>
		/// <param name="relativeOffset">The relative offset to move of the servo (Range: 0-255).</param>
		/// <param name="direction">The direction to move the servo.</param>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static void SetServoPositionRelative(byte iicAddress, ServoIdType servoId, byte relativeOffset, ServoRelativeDirectionType direction)
		{
			try
			{
				byte[] data = new byte[3] { (byte)servoId, relativeOffset, (byte)direction };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.PositionServoRelative, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.PositionServoRelative, Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Set a system parameter.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="systemParameter">The system parameter to set.</param>
		/// <param name="value">The new value for the parameter. The value type and contents must match
		/// the system parameter. (e.g. SystemParameterType.IicBaudRate value type is IicBaudRateType)</param>
		/// <exception cref="ArgumentOutOfRangeException">Thrown if the system parameter is readonly or if
		/// the specified value is out of range.</exception>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static void SetSystemParameter(byte iicAddress, SystemParameterType systemParameter, object value)
		{
			try
			{
				byte sendValue;

				switch (systemParameter)
				{
					case SystemParameterType.ModuleIicAddress:
					case SystemParameterType.RouterIicAddress:
						if ((value is byte) == false)
						{
							throw new ArgumentException(string.Format("The type {0} is invalid for the system parameter {1}.",
								value.GetType(), systemParameter), "value");
						}

						sendValue = (byte)value;
						if ((sendValue < 2) || (sendValue > 254) || (sendValue % 2 != 0))
						{
							throw new ArgumentOutOfRangeException("value", value, "Valid range between 2-254 (must be an even number).");
						}
						break;

					case SystemParameterType.HearbeatRate:
					case SystemParameterType.SerialTimeout:
						if ((value is byte) == false)
						{
							throw new ArgumentException(string.Format("The type {0} is invalid for the system parameter {1}.",
								value.GetType(), systemParameter), "value");
						}

						sendValue = (byte)value;
						if (sendValue < 1)
						{
							throw new ArgumentOutOfRangeException("value", value, "Valid range between 1-255.");
						}
						break;

					case SystemParameterType.IicBaudRate:
						if ((value is IicBaudRateType) == false)
						{
							throw new ArgumentException(string.Format("The type {0} is invalid for the system parameter {1}.",
								value.GetType(), systemParameter), "value");
						}

						sendValue = (byte)(IicBaudRateType)value;
						break;

					case SystemParameterType.SerialBaudRate:
						if ((value is SerialBaudRateType) == false)
						{
							throw new ArgumentException(string.Format("The type {0} is invalid for the system parameter {1}.",
								value.GetType(), systemParameter), "value");
						}

						sendValue = (byte)(SerialBaudRateType)value;
						break;

					case SystemParameterType.HeartbeatMode:
						if ((value is HeartbeatModeType) == false)
						{
							throw new ArgumentException(string.Format("The type {0} is invalid for the system parameter {1}.",
								value.GetType(), systemParameter), "value");
						}

						sendValue = (byte)(HeartbeatModeType)value;
						break;

					case SystemParameterType.BootReflexFlag:
						if ((value is BootReflexFlagType) == false)
						{
							throw new ArgumentException(string.Format("The type {0} is invalid for the system parameter {1}.",
								value.GetType(), systemParameter), "value");
						}

						sendValue = (byte)(BootReflexFlagType)value;
						break;

					case SystemParameterType.LinkDownReflexFlag:
						if ((value is LinkDownReflexFlagType) == false)
						{
							throw new ArgumentException(string.Format("The type {0} is invalid for the system parameter {1}.",
								value.GetType(), systemParameter), "value");
						}

						sendValue = (byte)(LinkDownReflexFlagType)value;
						break;

					case SystemParameterType.VmBootFile1:
					case SystemParameterType.VmBootFile2:
					case SystemParameterType.VmBootFile3:
					case SystemParameterType.VmBootFile4:
						//TODO
						throw new NotImplementedException();

					case SystemParameterType.SerialRelayFlag:
					case SystemParameterType.QuietModeFlag:
						if ((value is bool) == false)
						{
							throw new ArgumentException(string.Format("The type {0} is invalid for the system parameter {1}.",
								value.GetType(), systemParameter), "value");
						}

						sendValue = Convert.ToByte((bool)value);
						break;

					default:
						throw new ArgumentOutOfRangeException("systemParameter", systemParameter, "The specified system parameter is readonly.");
				}

				byte[] data = new byte[2] { (byte)systemParameter, sendValue };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.SetSystemParameter, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.SetSystemParameter, Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Deinitialize the command manager.
		/// </summary>
		public static void Shutdown()
		{
			sInstance = null;
			Instance.mIsInitialized = false;
		}

		/// <summary>
		/// Execute a software reset.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static void SoftwareReset(byte iicAddress)
		{
			try
			{
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.Reset, Packet.DirectionType.HostToStem);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.Reset, Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Stops a servo by setting the current position to the desired position.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="servoId">The ID of the servo to send the command to.</param>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static void StopServo(byte iicAddress, ServoIdType servoId)
		{
			try
			{
				byte[] data = new byte[1] { (byte)servoId };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.StopServo, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.StopServo, Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

		/// <summary>
		/// Write an output bit to a digital IO pin.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="digitalIoPin">The digital IO pin to send the output to.</param>
		/// <param name="value">The output to send to the digital IO pin.</param>
		/// <exception cref="CommandException">Thrown if there's an error processing the command.</exception>
		public static void WriteDigitalOutput(byte iicAddress, DigitalIoPinType digitalIoPin, bool value)
		{
			try
			{
				byte[] data = new byte[2] { (byte)digitalIoPin, Convert.ToByte((bool)value) };
				Packet packet = new Packet(iicAddress, Packet.CommandCodeType.DigitalIo, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.DigitalIo, Instance.mErrorQueue[0]);
				}
			}
			finally
			{
				Instance.ClearPacketQueues();
			}
		}

        #region Private Helpers
        /// <summary>
        /// Gets the singleton instance.
        /// </summary>
        /// <value>The singleton instance.</value>
        private static CommandManager Instance
        {
            get
            {
                if (sInstance == null)
                {
                    sInstance = new CommandManager();
                }

                return sInstance;
            }
        }

		/// <summary>
		/// Initializes a new instance of the CommandManager class.
		/// </summary>
		private CommandManager()
		{
			mPacketQueue = new PacketCollection();
			mMessageQueue = new PacketCollection();
			mErrorQueue = new PacketCollection();
		}

        /// <summary>
        /// Handle BrainStem heartbeat packets.
        /// </summary>
        /// <param name="sender">The sender of the event.</param>
        /// <param name="e">An EventArgs object that contains the event data.</param>
        private static void mStream_HeartbeatReceived(object sender, HeartbeatEventsArgs e)
        {
            HeartbeatType heartbeatType = HeartbeatType.HostToStemUp;
            bool sendReply;

            switch (e.HeartbeatType)
            {
                case HeartbeatType.StemToHostUp:
                    sendReply = true;
                    heartbeatType = HeartbeatType.HostToStemUp;
                    break;

                case HeartbeatType.SteamToHostDown:
                    sendReply = true;
                    heartbeatType = HeartbeatType.HostToStemDown;
                    break;

                default:
                    sendReply = false;
                    break;
            }

            if (sendReply)
            {
				byte[] data = new byte[1] { (byte)heartbeatType };
				Packet packet = new Packet(e.Packet.IicAddress, Packet.CommandCodeType.HeartBeat, Packet.DirectionType.HostToStem, data);
				Instance.mStream.WritePacket(packet);

				// Check the error queue.
				if (Instance.mErrorQueue.Count != 0)
				{
					throw new CommandException(Packet.CommandCodeType.HeartBeat, "Error processing Heartbeat.", Instance.mErrorQueue[0]);
				}
            }
        }

        /// <summary>
        /// Handle generic BrainStem packets.
        /// </summary>
        /// <param name="sender">The sender of the event.</param>
        /// <param name="e">An EventArgs object that contains the event data.</param>
        private static void mStream_PacketReceived(object sender, PacketEventsArgs e)
        {
			switch (e.Packet.CommandCode)
			{
				case Packet.CommandCodeType.Message:
					if (e.Packet.Data != null)
					{
						MessageCodeType messageCode = (MessageCodeType)e.Packet.Data[0];
						switch (messageCode)
						{
							case MessageCodeType.PowerUpReset:
							case MessageCodeType.VmExit:
								Instance.mMessageQueue.Add(e.Packet);
								break;

							default:
								Instance.mErrorQueue.Add(e.Packet);
								break;
						}
					}
					break;

				default:
					Instance.mPacketQueue.Add(e.Packet);
					break;
			}            
        }

        /// <summary>
        /// Retrieve a packet from the packet queue.
        /// </summary>
		/// <param name="commandCode">The <see cref="Packet.CommandCodeType"/> to locate in the packet queue.</param>
        /// <returns>The first packet in the queue that matches <paramref name="commandCode"/>. Null if no
		/// matching packet was found.</returns>
        private Packet GetPacketFromQueue(Packet.CommandCodeType commandCode)
        {
			int replyTime = 0;
			int index = Instance.mPacketQueue.IndexOf(commandCode);
			while (index == -1)
			{
				if (replyTime > Instance.mCommandReplyTimeout)
				{
					break;
				}

				replyTime += REPLY_SLEEP_TIME;
				Thread.Sleep(REPLY_SLEEP_TIME);
				index = Instance.mPacketQueue.IndexOf(commandCode);
			}

			if (index == -1)
			{
				return null;
			}
			else
			{
				return Instance.mPacketQueue[index];
			}
        }

		/// <summary>
		/// Clear each of the packet queues.
		/// </summary>
		private void ClearPacketQueues()
		{
			Instance.mPacketQueue.Clear();
			Instance.mMessageQueue.Clear();
			Instance.mErrorQueue.Clear();
		}

		/// <summary>
		/// Parse a Packet.CommandCodeType.DeviceInput value into it's component parts.
		/// </summary>
		/// <param name="deviceInputCode">The device input byte to parse.</param>
		/// <param name="deviceId">The device ID stored in the input code..</param>
		/// <param name="sendReplyToHost">The BrainStem will send a reply packet to the host.</param>
		/// <param name="sendReplyToReflex">The value will go directly to a reflex routine within the BrainStem.</param>
		private void ParseDeviceInputCode(byte deviceInputCode, out byte deviceId, out bool sendReplyToHost, out bool sendReplyToReflex)
		{
			byte deviceIDMask = 0x3F;
			sendReplyToHost = ((deviceInputCode & (byte)DeviceValueFlags.Host) != 0);
			sendReplyToReflex = ((deviceInputCode & (byte)DeviceValueFlags.Reflex) != 0);
			deviceId = (byte)(deviceInputCode & deviceIDMask);
		}
        #endregion

        private const int REPLY_SLEEP_TIME = 10;
		private const byte MAX_SERVO_POSITION_ZERO = 70;
		private const byte MIN_SERVO_POSITION_RESOLUTION = 1;
		private const byte MAX_SERVO_POSITION_RESOLUTION = 70;
		private const int MAX_A2D_VALUE = 1023;

        private static CommandManager sInstance;
        private Stream mStream;
		private PacketCollection mPacketQueue;
		private PacketCollection mErrorQueue;
		private PacketCollection mMessageQueue;
        private int mCommandReplyTimeout;
		private bool mIsInitialized;
    }
}
