using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using Bespoke.Robotics.BrainStem;

namespace Bespoke.Robotics.BrainStem.TestHarness
{
	/// <summary>
	/// Configuration form for a servo.
	/// </summary>
	public partial class ConfigServoForm : Form
	{
		/// <summary>
		/// Gets or sets the selected speed.
		/// </summary>
		/// <value>The selected speed.</value>
		/// <exception cref="ArgumentOutOfRangeException">Thrown if the property is set to a value out of
		/// range (range: 0-15).</exception>
		public int SelectedSpeed
		{
			get
			{
				return (int)speedComboBox.SelectedItem;
			}
			set
			{
				if ((value < MIN_SPEED) || (value > MAX_SPEED))
				{
					throw new ArgumentOutOfRangeException("value", value, string.Format("Valid range between {0}-{1}", MIN_SPEED, MAX_SPEED));
				}
				
				speedComboBox.SelectedIndex = value;
			}
		}

		/// <summary>
		/// Initializes a new instance of the ConfigServoForm class.
		/// </summary>
		/// <param name="iicAddress">The address of the IIC module referenced by the packet.</param>
		/// <param name="servoId"></param>
		public ConfigServoForm(byte iicAddress, ServoIdType servoId)
		{
			InitializeComponent();
			mModuleAddress = iicAddress;
			mServoId = servoId;
		}

		/// <summary>
		/// Initialize the form.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void ConfigServoForm_Load(object sender, EventArgs e)
		{
			try
			{
				byte positionZero, positionResolution;
				mServoConfiguration = CommandManager.GetServoConfiguration(mModuleAddress, mServoId);
				CommandManager.GetServoMotion(mModuleAddress, mServoId, out positionZero, out positionResolution);

				// Populate the speed combo box.
				for (int i = MIN_SPEED; i <= MAX_SPEED; i++)
				{
					speedComboBox.Items.Add(i);
				}

				Text = string.Format("Configure Servo {0}", (byte)mServoId);
				SelectedSpeed = mServoConfiguration.Speed;
				enabledCheckBox.Checked = mServoConfiguration.Enabled;
				reverseCheckBox.Checked = mServoConfiguration.Inverted;
				offsetTrackBar.Value = positionZero;
				rangeTrackBar.Value = positionResolution;
				mSendCommands = true;
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "Error Loading Servo Configuration");
				enabledCheckBox.Enabled = false;
				reverseCheckBox.Enabled = false;
				speedComboBox.Enabled = false;
				offsetTrackBar.Enabled = false;
				rangeTrackBar.Enabled = false;
				okButton.Enabled = false;
			}
		}

		/// <summary>
		/// Configure the offset and range of the servo.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void configTrackBar_Scroll(object sender, EventArgs e)
		{
			try
			{
				if (mSendCommands)
				{
					byte positionZero = (byte)offsetTrackBar.Value;
					byte positionResolution = (byte)rangeTrackBar.Value;

					CommandManager.ConfigureServoMotion(mModuleAddress, mServoId, positionZero, positionResolution);
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "Error Setting Servo Motion");
			}
		}

		/// <summary>
		/// Configure the speed, enabled and reverse states of the servo.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void checkBox_CheckedChanged(object sender, EventArgs e)
		{
			try
			{
				if (mSendCommands)
				{
					CommandManager.ConfigureServo(mModuleAddress, new ServoConfiguration(mServoId, enabledCheckBox.Checked, reverseCheckBox.Checked, (byte)SelectedSpeed));
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "Error Setting Servo Configuration");
			}
		}

		private const int MIN_SPEED = 0;
		private const int MAX_SPEED = 15;

		private bool mSendCommands;
		private byte mModuleAddress;
		private ServoIdType mServoId;
		private ServoConfiguration mServoConfiguration;
	}
}