using System;
using System.Diagnostics;
using System.Threading;
using System.Reflection;
using System.ComponentModel;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Bespoke.Robotics.Common;
using Bespoke.Robotics.BrainStem;
using Bespoke.Robotics.BrainStem.Configuration;

namespace Bespoke.Robotics.BrainStem.TestHarness
{
	/// <summary>
	/// Main UI for the BrainStem test harness.
	/// </summary>
    public partial class MainForm : Form
    {
		/// <summary>
		/// Initializes a new instance of the MainForm class.
		/// </summary>
        public MainForm()
        {
            InitializeComponent();

			try
			{
				mModuleAddress = Configuration.Configuration.BrainStem.ModuleAddress;
				mStream = new Stream();
				mStream.HeartbeatReceived += new HeartbeatReceivedEventHandler(mStream_HeartbeatReceived);
				mStream.Open();
				CommandManager.Init(mStream, Configuration.Configuration.BrainStem.CommandReplyTimeout);
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message);
			}
		}

		#region Private Helpers
		/// <summary>
		/// Initialize the list of system parameters that can be set or retrieved.
		/// </summary>
		private void InitSystemParamComboBox()
		{
			systemParamComboBox.Items.Clear();
			foreach (FieldInfo fieldInfo in typeof(SystemParameterType).GetFields())
			{
				if (fieldInfo.Name != "value__")
				{
					systemParamComboBox.Items.Add(fieldInfo.Name);
				}
			}

			systemParamValueTextBox.Clear();
		}

		/// <summary>
		/// Initialize the servo positions.
		/// </summary>
		private void InitServoPositions()
		{
			servo0UpDown.Value = CommandManager.GetServoPosition(mModuleAddress, ServoIdType.Servo0);
			servo1UpDown.Value = CommandManager.GetServoPosition(mModuleAddress, ServoIdType.Servo1);
			servo2UpDown.Value = CommandManager.GetServoPosition(mModuleAddress, ServoIdType.Servo2);
			servo3UpDown.Value = CommandManager.GetServoPosition(mModuleAddress, ServoIdType.Servo3);
		}

		/// <summary>
		/// Initialize the A2D inputs.
		/// </summary>
		private void InitA2DValues()
		{
			try
			{
				analog0ProgressBar.Value = CommandManager.ReadA2DInput(mModuleAddress, AnalogToDigitalIoPinType.Pin0, true, false);
				analog1ProgressBar.Value = CommandManager.ReadA2DInput(mModuleAddress, AnalogToDigitalIoPinType.Pin1, true, false);
				analog2ProgressBar.Value = CommandManager.ReadA2DInput(mModuleAddress, AnalogToDigitalIoPinType.Pin2, true, false);
				analog3ProgressBar.Value = CommandManager.ReadA2DInput(mModuleAddress, AnalogToDigitalIoPinType.Pin3, true, false);
				analog4ProgressBar.Value = CommandManager.ReadA2DInput(mModuleAddress, AnalogToDigitalIoPinType.Pin4, true, false);				
			}
			catch (Exception ex)
			{
				analogTimer.Enabled = false;
				MessageBox.Show(ex.Message, "Error Reading A2D Input");
			}
		}

		/// <summary>
		/// Update the heart beat UI.
		/// </summary>
		/// <param name="up">The type of heart beat to set (either up or down).</param>
		private void SetHeartBeatButton(bool up)
		{
			try
			{
				if (heartbeatRadioButton.InvokeRequired)
				{
					SetHeartbeatButtonHandler handler = new SetHeartbeatButtonHandler(SetHeartBeatButton);
					BeginInvoke(handler, new object[] { up });
				}
				else
				{
					heartbeatRadioButton.Checked = up;
				}
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex.Message);
			}
		}
		#endregion

		#region Event Handlers
		/// <summary>
		/// Initialize the form.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An EventArgs object containing the event data.</param>
		private void MainForm_Load(object sender, EventArgs e)
		{
			try
			{
				InitSystemParamComboBox();
				mSendCommands = true;
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "Error Loading Test Harness");
				tabControl.Enabled = false;
			}
		}

		/// <summary>
		/// Initialize test sections as tab pages change.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void tabControl_SelectedIndexChanged(object sender, EventArgs e)
		{
			TabPageTags tabPageTag = (TabPageTags)Enum.Parse(typeof(TabPageTags), (string)tabControl.SelectedTab.Tag);
			analogTimer.Enabled = false;

			try
			{
				switch (tabPageTag)
				{
					case TabPageTags.Servo:
						InitServoPositions();
						break;

					case TabPageTags.Analog:
						analogTimer.Enabled = true;
						break;
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, string.Format("Error Initializing Tab Page: {0}", tabPageTag.ToString()));
			}
		}

		/// <summary>
		/// Handle the BrainStem Heartbeat event.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
        private void mStream_HeartbeatReceived(object sender, HeartbeatEventsArgs e)
        {
			bool heartbeatUp = (e.HeartbeatType == HeartbeatType.StemToHostUp);
			SetHeartBeatButton(heartbeatUp);
        }

		/// <summary>
		/// Software reset.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
        private void resetButton_Click(object sender, EventArgs e)
        {
			Cursor = Cursors.WaitCursor;

			try
			{
				if (mSendCommands)
				{
					CommandManager.SoftwareReset(mModuleAddress);
					Thread.Sleep(1000);

					heartbeatModeButton.Text = INTERNAL_HEARTBEAT_TEXT;
					debugTextBox.Text = String.Empty;
					outputTextBox.Text = String.Empty;
					InitSystemParamComboBox();
					InitServoPositions();
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "Error Performing Software Reset");
			}

			Cursor = Cursors.Default;
        }

		/// <summary>
		/// Close the form.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
        private void closeButton_Click(object sender, EventArgs e)
        {
            mStream.Close();
            Close();
        }

		/// <summary>
		/// Debug (echo) commands.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void debugButton_Click(object sender, EventArgs e)
		{
			try
			{
				if (mSendCommands)
				{
					byte[] data = null;

					if (debugTextBox.Text != String.Empty)
					{
						string[] debugCharacters = debugTextBox.Text.Split(" ".ToCharArray(), StringSplitOptions.RemoveEmptyEntries);
						int length = (debugCharacters.Length > Packet.MAX_DATA_BYTES ? Packet.MAX_DATA_BYTES : debugCharacters.Length);
						data = new byte[length];
						for (int i = 0; i < length; i++)
						{
							data[i] = Convert.ToByte(debugCharacters[i]);
						}
					}

					byte[] replyData = CommandManager.Debug(mModuleAddress, data);
					if (replyData != null)
					{
						StringBuilder replyString = new StringBuilder();
						foreach (byte replyChar in replyData)
						{
							replyString.Append(replyChar.ToString());
							replyString.Append(" ");
						}

						outputTextBox.Text = replyString.ToString() + Environment.NewLine + outputTextBox.Text;
					}
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "Error Sending Debug Command");
			}
		}

		/// <summary>
		/// Manage servo position through track bar movement.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void servoTrackBar_ValueChanged(object sender, EventArgs e)
		{
			try
			{
				TrackBar trackBar = (TrackBar)sender;
				ServoIdType servoId = (ServoIdType)Enum.Parse(typeof(ServoIdType), (string)trackBar.Tag);

				if (mSendCommands)
				{
					CommandManager.SetServoPositionAbsolute(mModuleAddress, servoId, (byte)trackBar.Value);
				}

				NumericUpDown upDown = null;
				switch (servoId)
				{
					case ServoIdType.Servo0:
						upDown = servo0UpDown;
						break;

					case ServoIdType.Servo1:
						upDown = servo1UpDown;
						break;

					case ServoIdType.Servo2:
						upDown = servo2UpDown;
						break;

					case ServoIdType.Servo3:
						upDown = servo3UpDown;
						break;

					default:
						throw new Exception("Unknown Servo ID");
				}

				upDown.Value = trackBar.Value;
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "Error Setting Servo Position");
			}
		}

		/// <summary>
		/// Manage servo position through servo numeric up/down movement.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void servoUpDown_ValueChanged(object sender, EventArgs e)
		{
			NumericUpDown upDown = (NumericUpDown)sender;
			ServoIdType servoId = (ServoIdType)Enum.Parse(typeof(ServoIdType), (string)upDown.Tag);
			TrackBar trackBar = null;
			switch (servoId)
			{
				case ServoIdType.Servo0:
					trackBar = servo0TrackBar;
					break;

				case ServoIdType.Servo1:
					trackBar = servo1TrackBar;
					break;

				case ServoIdType.Servo2:
					trackBar = servo2TrackBar;
					break;

				case ServoIdType.Servo3:
					trackBar = servo3TrackBar;
					break;

				default:
					throw new Exception("Unknown Servo ID");
			}

			if (trackBar.Value != (int)upDown.Value)
			{
				trackBar.Value = (int)upDown.Value;
			}
		}

		/// <summary>
		/// Configure servos.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void servoConfigButton_Click(object sender, EventArgs e)
		{
			Button configButton = ((Button)sender);
			ServoIdType servoId = (ServoIdType)Enum.Parse(typeof(ServoIdType), (string)configButton.Tag);
			ConfigServoForm configServoForm = new ConfigServoForm(mModuleAddress, servoId);
			configServoForm.ShowDialog();
		}

		/// <summary>
		/// Commit servo settings to the BrainStem EEPROM.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void commitServoSettingsButton_Click(object sender, EventArgs e)
		{
			try
			{
				if (mSendCommands)
				{
					CommandManager.SaveServoParameters(mModuleAddress);
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "Error Committing Servo Settings");
			}
		}

		/// <summary>
		/// Change the heartbeat mode internal/external.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void heartbeatModeButton_Click(object sender, EventArgs e)
		{
			try
			{
				if (mSendCommands)
				{
					HeartbeatModeType newMode = (heartbeatModeButton.Text == INTERNAL_HEARTBEAT_TEXT ? HeartbeatModeType.Internal : HeartbeatModeType.LinkDependent);
					CommandManager.SetSystemParameter(mModuleAddress, SystemParameterType.HeartbeatMode, newMode);
					heartbeatModeButton.Text = (newMode == HeartbeatModeType.Internal ? EXTERNAL_HEARTBEAT_TEXT : INTERNAL_HEARTBEAT_TEXT);
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "Error Setting Heartbeat Mode");
			}
		}

		/// <summary>
		/// Clear the get/set parameter value.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void systemParamComboBox_SelectedIndexChanged(object sender, EventArgs e)
		{
			systemParamValueTextBox.Clear();
		}

		/// <summary>
		/// Get a BrainStem system parameter.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void systemParamGetValueButton_Click(object sender, EventArgs e)
		{
			try
			{
				if (mSendCommands)
				{
					SystemParameterType systemParameter = (SystemParameterType)Enum.Parse(typeof(SystemParameterType), (string)systemParamComboBox.SelectedItem);
					string value = CommandManager.GetSystemParameter(mModuleAddress, systemParameter);
					systemParamValueTextBox.Text = value;
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message, "Error Retrieving System Parameters");
			}
		}

		/// <summary>
		/// Read the A2D values.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">An <see "EventArgs"> object containing the event data.</param>
		private void analogTimer_Tick(object sender, EventArgs e)
		{
			InitA2DValues();
		}
		#endregion

		private delegate void SetHeartbeatButtonHandler(bool up);

		private enum TabPageTags
		{
			Debug			= 0,
			SystemParams,
			Servo,
			Analog
		}

		private const string INTERNAL_HEARTBEAT_TEXT = "Use Internal Heartbeat";
		private const string EXTERNAL_HEARTBEAT_TEXT = "Use Link-Dependent Heartbeat";

        private byte mModuleAddress;
        private Stream mStream;
		private bool mSendCommands;
    }
}